package dynaaminen.stringi;

import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 * Esimerkki dynaamisesta taulukosta
 * @author Vesa Lappalainen @version 1.3 25.2.2019
 * @param <TYPE> minkä tyyppinen taulukko tehdään
 * @example
 */
public class Taulukko<TYPE> implements Iterable<TYPE>{
    /** Luokka täyden taulukon poikkeusta varten  */
    public static class TaulukkoTaysiException extends Exception {
        private static final long serialVersionUID = 1L;

        TaulukkoTaysiException(String viesti) {
            super(viesti);
        }
    }

    private int lkm;
    private TYPE[] alkiot;
    
    /**
     * ALustetaan taulukko 
     * @param maxKoko alkukoko
     */
    @SuppressWarnings("unchecked")
    public Taulukko(int maxKoko) {
        lkm = 0;
        alkiot = (TYPE[])new Object[maxKoko];
    }
    
    
    /**
     * Lisää uusi alkio
     * @param uusi lisättävä alkio
     * @throws TaulukkoTaysiException jos tila loppuu 
     */
    public void lisaa(TYPE uusi) throws TaulukkoTaysiException {
        if ( lkm >= alkiot.length ) throw new TaulukkoTaysiException("Tila loppu!");
        alkiot[lkm] = uusi;
        lkm++;
    }
    
    
    @Override
    public String toString() {
        StringBuilder tulos = new StringBuilder();
        String ero = "";
        for (int i = 0; i < lkm; i++) {
            TYPE alkio = alkiot[i];
            tulos.append(ero);
            tulos.append(alkio);
            ero = ",";
        }
        return tulos.toString();
    }
    
    
    /**
     * @param i  mihin paikkaan
     * @param uusi mikä arvo
     */
    public void set(int i, TYPE uusi) {
        if ( lkm <= i ) throw new IndexOutOfBoundsException("Laiton indeksi set: " + i);
        alkiot[i] = uusi;
    }
    
    
    /**
     * @param i mistä paikasta
     * @return paikassa oeva arvo
     */
    public TYPE get(int i) {
        if ( i < 0 || lkm <= i ) throw new IndexOutOfBoundsException("Laiton indeksi: " + i);
        return alkiot[i];   
    }
    
    /**
     * Iteraattori lukujen läpikäymiseksi
     */
    public class Iter implements Iterator<TYPE> {
        int kohdalla;


        /** Alustetaan taulukon ulkopuolelle  */
        public Iter() {
            kohdalla = 0;
        }


        @Override
        public boolean hasNext() {
            return kohdalla < lkm;
        }


        @Override
        public TYPE next() {
            if (!hasNext()) throw new NoSuchElementException("Ei oo enää");
            return get(kohdalla++);
        }


        @Override
        public void remove() {
            throw new UnsupportedOperationException("Ei voi poistaa");      
        }        
    }
    
    
    /**
     * @return iteraattori, jolla taulukkoa voidaan käydä lävitse.
     * @example
     * <pre name="test">
     *  #THROWS TaulukkoIter.TaulukkoTaysiException
     *  #import java.util.*;
     *   TaulukkoIter t = new TaulukkoIter();
     *   t.lisaa(2); t.lisaa(99); t.lisaa(5);
     *   int[] luvut = {2,99,5};
     *   int i=0;
     *   for (Iterator<Integer> iter = t.iterator(); iter.hasNext(); )
     *      iter.next() === new Integer(luvut[i++]);
     *   i=0;
     *   for (int luku:t)
     *      luku === luvut[i++];
     * </pre>
     */
    @Override
    public Iterator<TYPE> iterator() {
        return new Iter();
    }
    
    
    /**
     * kokeillaan taulukkoluokkaa 
     * @param args ei käytössä
     */
    public static void main(String[] args) {
        Taulukko<String> luvut = new Taulukko<String>(8);
        System.out.println(luvut.toString());
        try {
            luvut.lisaa("Kissa");
            luvut.lisaa("Kana");
            System.out.println(luvut.toString());
            luvut.lisaa("Kettu");
            System.out.println(luvut.toString());
            luvut.lisaa("Koira");
            luvut.set(1,"Mato");

            for (var iter=luvut.iterator(); iter.hasNext(); ) {
                var alkio = iter.next();
                System.out.print(alkio + " ");
            }        
            System.out.println();

            for (var alkio : luvut) {
                System.out.print(alkio + " ");
            }        
            System.out.println();
            System.out.println(luvut);
        } catch (IndexOutOfBoundsException e) {
            // e.printStackTrace();
            System.err.println("Tapahtui virhe: " + e.getMessage());
        } catch (TaulukkoTaysiException e) {
            // e.printStackTrace();
            System.err.println("Tapahtui virhe: " + e.getMessage());
        }
    }
}
